/**
 * @namespace   biew
 * @file        setup.h
 * @brief       This file included BIEW setup functions description.
 * @version     -
 * @remark      this source file is part of Binary vIEW project (BIEW).
 *              The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.
 *              All rights reserved. This software is redistributable under the
 *              licence given in the file "Licence.en" ("Licence.ru" in russian
 *              translation) distributed in the BIEW archive.
 * @note        Requires POSIX compatible development system
 *
 * @author      Nick Kurshev
 * @since       1999
 * @note        Development, fixes and improvements
**/
#ifndef __SETUP__H
#define __SETUP__H

#ifdef __cplusplus
extern "C" {
#endif

extern char *     biewGetHelpName( void );
extern void       Setup( void );

#ifdef __cplusplus
}
#endif

#endif
