/**
 * @namespace   biewlib
 * @file        biewlib/sysdep/ia16/__config.h
 * @brief       This file configures project for 16-bit Intel architecture.
 * @version     -
 * @remark      this source file is part of Binary vIEW project (BIEW).
 *              The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.
 *              All rights reserved. This software is redistributable under the
 *              licence given in the file "Licence.en" ("Licence.ru" in russian
 *              translation) distributed in the BIEW archive.
 * @note        Requires POSIX compatible development system
 *
 * @author      Nick Kurshev
 * @since       2000
 * @note        Development, fixes and improvements
 * @warning     We must have own version of included files for posibility to build project under any development systems.
**/
#ifndef ____CONFIG_H
#define ____CONFIG_H 1

#include "biewlib/sysdep/ia16/endian.h"
#include "biewlib/sysdep/ia16/stdint.h"

#endif
